@@ -1,4 +1,4 @@
-#!/usr/bin/python2
+#!/usr/bin/python3
 
 # Todo:
 # - TextTag.invisible does not work nicely with scrollheight, find out why
@@ -15,9 +15,9 @@ import os
 import pyinotify
 import gi
 gi.require_version('Gtk', '3.0')
-from gi.repository import Gio, Gtk, GObject, Gdk, Pango, GLib
+from gi.repository import Gio, Gtk, GObject, Gdk, GLib
 import dbus, dbus.service, dbus.glib
-from pageutils import *
+import pageutils
 from lookingglass_proxy import LookingGlassProxy
 from dbus.mainloop.glib import DBusGMainLoop
 import signal
@@ -37,7 +37,7 @@ class MenuButton(Gtk.Button):
     def onClicked(self, widget):
         x, y, w, h = self.getScreenCoordinates()
         self.menu.popup(None, None, lambda menu, data: (x, y+h, True), None, 1, 0)
-                
+
     def getScreenCoordinates(self):
         parent = self.get_parent_window()
         x, y = parent.get_root_origin()
@@ -45,73 +45,11 @@ class MenuButton(Gtk.Button):
         h = parent.get_height()
         extents = parent.get_frame_extents()
         allocation = self.get_allocation()
-        return (x + (extents.width-w)/2 + allocation.x,
-                y + (extents.height-h)-(extents.width-w)/2 + allocation.y,
+        return (x + (extents.width-w)//2 + allocation.x,
+                y + (extents.height-h)-(extents.width-w)//2 + allocation.y,
                 allocation.width,
                 allocation.height)
 
-class ResizeGrip(Gtk.Widget):
-    def __init__(self, parent):
-        Gtk.Widget.__init__(self)
-        self.parentWindow = parent
-
-    def do_realize(self):
-        self.set_realized(True)
-
-        allocation = self.get_allocation()
-
-        attr = Gdk.WindowAttr()
-        attr.window_type = Gdk.WindowType.CHILD
-        attr.wclass = Gdk.WindowWindowClass.INPUT_OUTPUT
-        attr.event_mask = self.get_events() | Gdk.EventMask.EXPOSURE_MASK | Gdk.EventMask.BUTTON_PRESS_MASK
-        attr.x = 0
-        attr.y = 0
-        attr.width = allocation.width
-        attr.height = allocation.height
-
-        mask = Gdk.WindowAttributesType.X | Gdk.WindowAttributesType.Y
-        window = Gdk.Window.new(self.get_parent_window(), attr, mask)
-        self.set_window(window)
-
-        window.set_user_data(self)
-        self.style = self.get_style()
-        self.style.set_background(window, Gtk.StateFlags.NORMAL)
-
-        self.get_window().set_cursor(Gdk.Cursor(Gdk.CursorType.BOTTOM_SIDE))
-        self.connect("draw", self.onDraw)
-
-    def do_unrealize(self):
-        self.get_window().destroy()
-
-    def do_size_request(self, requisition):
-        requisition.height = 4
-        requisition.width = -1
-
-    def do_get_preferred_width(self):
-        req = Gtk.Requisition()
-        self.do_size_request(req)
-        return (req.width, req.width)
-
-    def do_get_preferred_height(self):
-        req = Gtk.Requisition()
-        self.do_size_request(req)
-        return (req.height, req.height)
-
-    def do_size_allocate(self, allocation):
-        if self.get_realized():
-            self.get_window().move_resize(allocation.x, allocation.y, allocation.width, allocation.height)
-            self.queue_draw()
-
-    def do_button_press_event(self, event):
-        self.parentWindow.begin_resize_drag(Gdk.WindowEdge.SOUTH, event.button, int(event.x_root), int(event.y_root), event.time)
-        return True
-
-    def onDraw(self, widget, ctx):
-        width = self.get_window().get_width()
-        # Draw a line using the current theme
-        cr = self.get_window().cairo_create()
-        self.style.do_draw_hline(self.style, cr, Gtk.StateType.NORMAL, self, "", 1, width-2, 1)
-
 class CommandLine(Gtk.Entry):
     def __init__(self):
         Gtk.Entry.__init__(self)
@@ -137,7 +75,7 @@ class CommandLine(Gtk.Entry):
         if event.keyval == Gdk.KEY_Down:
             self.historyNext()
             return True
-        if event.keyval == Gdk.KEY_Return:
+        if event.keyval == Gdk.KEY_Return or event.keyval == Gdk.KEY_KP_Enter:
             self.execute()
             return True
 
@@ -190,8 +128,8 @@ class CommandLine(Gtk.Entry):
 class NewLogDialog(Gtk.Dialog):
     def __init__(self, parent):
         Gtk.Dialog.__init__(self, "Add a new file watcher", parent, 0,
-            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
-             Gtk.STOCK_OK, Gtk.ResponseType.OK))
+                            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
+                             Gtk.STOCK_OK, Gtk.ResponseType.OK))
 
         self.set_default_size(150, 100)
 
@@ -200,7 +138,7 @@ class NewLogDialog(Gtk.Dialog):
 
         box = self.get_content_area()
         box.add(label)
-        
+
         self.store = Gtk.ListStore(str, str)
         self.store.append(["glass.log", "~/.cinnamon/glass.log"])
         self.store.append(["custom", "<Select file>"])
@@ -210,18 +148,18 @@ class NewLogDialog(Gtk.Dialog):
         renderer_text = Gtk.CellRendererText()
         self.combo.pack_start(renderer_text, True)
         self.combo.add_attribute(renderer_text, "text", 1)
-        
+
         table = Gtk.Table(2, 2, False)
         table.attach(Gtk.Label(label="File: ", halign=Gtk.Align.START), 0, 1, 0, 1)
         table.attach(self.combo, 1, 2, 0, 1)
         table.attach(Gtk.Label(label="Name: ", halign=Gtk.Align.START), 0, 1, 1, 2)
         self.entry = Gtk.Entry()
         table.attach(self.entry, 1, 2, 1, 2)
-        
+
         self.filename = None
         box.add(table)
         self.show_all()
-        
+
     def onComboChanged(self, combo):
         tree_iter = combo.get_active_iter()
         if tree_iter != None:
@@ -235,21 +173,21 @@ class NewLogDialog(Gtk.Dialog):
                 else:
                     combo.set_active(-1)
             return False
-                    
+
     def isValid(self):
         return self.entry.get_text() != "" and self.filename != None and os.path.isfile(os.path.expanduser(self.filename))
-        
+
     def getFile(self):
         return os.path.expanduser(self.filename)
-        
+
     def getName(self):
         return self.entry.get_text()
-        
+
     def selectFile(self):
         dialog = Gtk.FileChooserDialog("Please select a log file", self,
-            Gtk.FileChooserAction.OPEN,
-            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
-             Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
+                                       Gtk.FileChooserAction.OPEN,
+                                       (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
+                                        Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
 
         filter_text = Gtk.FileFilter()
         filter_text.set_name("Text files")
@@ -260,19 +198,19 @@ class NewLogDialog(Gtk.Dialog):
         filter_any.set_name("Any files")
         filter_any.add_pattern("*")
         dialog.add_filter(filter_any)
-        
+
         response = dialog.run()
         result = None
         if response == Gtk.ResponseType.OK:
             result = dialog.get_filename()
         dialog.destroy()
-        
+
         return result
 
 class FileWatchHandler(pyinotify.ProcessEvent):
     def __init__(self, view):
         self.view = view
-        
+
     def process_IN_CLOSE_WRITE(self, event):
         self.view.getUpdates()
 
@@ -291,6 +229,7 @@ class FileWatcherView(Gtk.ScrolledWindow):
 
         self.filename = filename
         self.changed = 0
+        self.updateId = 0
         self.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
         self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
 
@@ -302,7 +241,7 @@ class FileWatcherView(Gtk.ScrolledWindow):
 
         self.show_all()
         self.getUpdates()
-        
+
         handler = FileWatchHandler(self)
         wm = pyinotify.WatchManager()
         self.notifier = pyinotify.ThreadedNotifier(wm, handler)
@@ -310,12 +249,12 @@ class FileWatcherView(Gtk.ScrolledWindow):
         self.notifier.start()
         self.connect("destroy", self.onDestroy)
         self.connect("size-allocate", self.onSizeChanged)
-        
+
     def onDestroy(self, widget):
         if self.notifier:
             self.notifier.stop()
             self.notifier = None
-        
+
     def onSizeChanged(self, widget, bla):
         if self.changed > 0:
             end_iter = self.textbuffer.get_end_iter()
@@ -323,9 +262,17 @@ class FileWatcherView(Gtk.ScrolledWindow):
             self.changed -= 1
 
     def getUpdates(self):
+        # only update 2 times per second max
+        # without this rate limiting, certain file modifications can cause a crash at Gtk.TextBuffer.set_text()
+        if self.updateId == 0:
+            self.updateId = GLib.timeout_add(500, self.update)
+
+    def update(self):
         self.changed = 2 # onSizeChanged will be called twice, but only the second time is final
         self.textbuffer.set_text(open(self.filename, 'r').read())
-                
+        self.updateId = 0
+        return False
+
 class ClosableTabLabel(Gtk.Box):
     __gsignals__ = {
         "close-clicked": (GObject.SIGNAL_RUN_FIRST, GObject.TYPE_NONE, ()),
@@ -334,34 +281,23 @@ class ClosableTabLabel(Gtk.Box):
         Gtk.Box.__init__(self)
         self.set_orientation(Gtk.Orientation.HORIZONTAL)
         self.set_spacing(5)
-        
+
         label = Gtk.Label(label_text)
         self.pack_start(label, True, True, 0)
-        
+
         button = Gtk.Button()
         button.set_relief(Gtk.ReliefStyle.NONE)
         button.set_focus_on_click(False)
         button.add(Gtk.Image.new_from_stock(Gtk.STOCK_CLOSE, Gtk.IconSize.MENU))
         button.connect("clicked", self.button_clicked)
-        data =  ".button {\n" \
-                "-GtkButton-default-border : 0px;\n" \
-                "-GtkButton-default-outside-border : 0px;\n" \
-                "-GtkButton-inner-border: 0px;\n" \
-                "-GtkWidget-focus-line-width : 0px;\n" \
-                "-GtkWidget-focus-padding : 0px;\n" \
-                "padding: 0px;\n" \
-                "}"
-        provider = Gtk.CssProvider()
-        provider.load_from_data(data)
-        button.get_style_context().add_provider(provider, 600) 
         self.pack_start(button, False, False, 0)
-        
+
         self.show_all()
-    
+
     def button_clicked(self, button, data=None):
         self.emit("close-clicked")
 
-class CinnamonLog(dbus.service.Object):
+class MelangeApp(dbus.service.Object):
     def __init__ (self):
         global lookingGlassProxy
         lookingGlassProxy = LookingGlassProxy()
@@ -377,7 +313,10 @@ class CinnamonLog(dbus.service.Object):
     @dbus.service.method (MELANGE_DBUS_NAME, in_signature='', out_signature='')
     def show(self):
         if self.window.get_visible():
-            self.window.hide()
+            if self._minimized:
+                self.window.present()
+            else:
+                self.window.hide()
         else:
             self.showAndFocus()
 
@@ -396,36 +335,37 @@ class CinnamonLog(dbus.service.Object):
             self.window.hide()
 
     def showAndFocus(self):
-        self.window.present()
         self.window.show_all()
         self.lookingGlassProxy.refreshStatus()
-        screen = self.window.get_screen()
-        geom = screen.get_monitor_geometry(screen.get_primary_monitor())
-        w, h = self.window.get_size()
-        self.window.resize(geom.width, h)
-        self.window.move(geom.x, geom.y)
         self.commandline.grab_focus()
 
     def run(self):
         self.window = Gtk.Window(Gtk.WindowType.TOPLEVEL)
-        self.window.set_type_hint(Gdk.WindowTypeHint.UTILITY)
-        screen = self.window.get_screen()
-        geom = screen.get_monitor_geometry(screen.get_primary_monitor())
+        self.window.set_title("Melange")
+        self.window.set_icon_name("system-search")
+        self.window.set_default_size(1000, 400)
+        self.window.set_position(Gtk.WindowPosition.MOUSE)
 
-        self.window.set_border_width(0)
-        self.window.set_decorated(False)
-        self.window.set_skip_taskbar_hint(True)
+        # I can't think of a way to reliably detect if the window
+        # is active to determine if we need to present or hide
+        # in show(). Since the window briefly loses focus during
+        # shortcut press we'd be unable to detect it at that time.
+        # Keeping the window on top ensures the window is never
+        # obscured so we can just hide if visible.
         self.window.set_keep_above(True)
-        self.window.set_default_size(geom.width, 200 * self.window.get_scale_factor())
-        self.window.move(geom.x,geom.y)
-        self.window.stick()
 
         self.window.connect("delete_event", self.onDelete)
         self.window.connect("key-press-event", self.onKeyPress)
+        self._minimized = False
+        self.window.connect("window-state-event", self.onWindowState)
 
         numRows = 3
         numColumns = 6
         table = Gtk.Table(numRows, numColumns, False)
+        table.set_margin_start(6)
+        table.set_margin_end(6)
+        table.set_margin_top(6)
+        table.set_margin_bottom(6)
         self.window.add(table)
 
         self.notebook = Gtk.Notebook()
@@ -440,9 +380,10 @@ class CinnamonLog(dbus.service.Object):
         self.notebook.set_action_widget(label, Gtk.PackType.END)
 
         self.pages = {}
+        self.customPages = {}
         self.createPage("Results", "results")
         self.createPage("Inspect", "inspect")
-        self.createPage("Memory", "memory")
+        # self.createPage("Memory", "memory") - TODO: re-implement get_memory_info from cjs
         self.createPage("Windows", "windows")
         self.createPage("Extensions", "extensions")
         self.createPage("Log", "log")
@@ -450,15 +391,21 @@ class CinnamonLog(dbus.service.Object):
         table.attach(self.notebook, 0, numColumns, 0, 1)
 
         column = 0
-        pickerButton = ImageButton("gtk-color-picker", Gtk.IconSize.SMALL_TOOLBAR)
+        pickerButton = pageutils.ImageButton("color-select-symbolic")
+        pickerButton.set_tooltip_text("Select an actor to inspect")
         pickerButton.connect("clicked", self.onPickerClicked)
         table.attach(pickerButton, column, column+1, 1, 2, 0, 0, 2)
         column += 1
 
-        table.attach(Gtk.Label("Exec:"), column, column+1, 1, 2, 0, 0, 3)
+        fullGc = pageutils.ImageButton("user-trash-full-symbolic")
+        fullGc.set_tooltip_text("Invoke garbage collection")
+        # ignore signal arg
+        fullGc.connect ('clicked', lambda source: lookingGlassProxy.FullGc())
+        table.attach(fullGc, column, column+1, 1, 2, 0, 0, 2)
         column += 1
 
         self.commandline = CommandLine()
+        self.commandline.set_tooltip_text("Evaluate javascript")
         table.attach(self.commandline, column, column+1, 1, 2, Gtk.AttachOptions.EXPAND|Gtk.AttachOptions.FILL, 0, 3, 2)
         column += 1
 
@@ -469,33 +416,25 @@ class CinnamonLog(dbus.service.Object):
         table.attach(self.statusLabel, column, column+1, 1, 2, 0, 0, 1)
         column += 1
 
+        box = Gtk.HBox()
         settings = Gio.Settings("org.cinnamon.desktop.keybindings")
-        arr = settings.get_strv("looking-glass-keybinding");
-        accel = ""
-        done_one = False;
-
-        for element in arr:
-            if done_one:
-                accel += ", "
-
-            accel += element.replace("<", "&lt;").replace(">", "&gt;")
-            if not done_one:
-                done_one = True
-
-        keybinding = Gtk.Label()
-        keybinding.set_markup('<i>Toggle shortcut: %s</i>' % accel)
+        arr = settings.get_strv("looking-glass-keybinding")
+        if len(arr) > 0:
+            # only the first mapped keybinding
+            [accelKey, mask] = Gtk.accelerator_parse(arr[0])
+            if accelKey == 0 and mask == 0:
+                # failed to parse, fallback to plain accel string
+                label = Gtk.Label(arr[0])
+            else:
+                label = Gtk.Label(Gtk.accelerator_get_label(accelKey, mask))
+            label.set_tooltip_text("Toggle shortcut")
+            box.pack_start(label, False, False, 3)
 
         actionButton = self.createActionButton()
-
-        box = Gtk.HBox()
-        box.pack_start(keybinding, False, False, 3)
         box.pack_start(actionButton, False, False, 3)
 
         table.attach(box, column, column+1, 1, 2, 0, 0, 1)
 
-        grip = ResizeGrip(self.window)
-        table.attach(grip, 0, numColumns, 2, 3, Gtk.AttachOptions.EXPAND|Gtk.AttachOptions.FILL, 0, 0, 0)
-
         self.activatePage("results")
         self.statusLabel.hide()
         self.window.set_focus(self.commandline)
@@ -514,10 +453,10 @@ class CinnamonLog(dbus.service.Object):
         menu.append(self.createMenuItem('Reset Cinnamon Settings', self.onResetClicked))
         menu.append(Gtk.SeparatorMenuItem())
         menu.append(self.createMenuItem('About Melange', self.onAboutClicked))
-        menu.append(self.createMenuItem('Quit', lambda x: Gtk.main_quit()))
+        menu.append(self.createMenuItem('Quit', self.onDelete))
         menu.show_all()
 
-        button = Gtk.MenuButton(u"Actions \u25BE")
+        button = Gtk.MenuButton("Actions \u25BE")
         button.set_popup(menu)
         return button
 
@@ -530,18 +469,20 @@ class CinnamonLog(dbus.service.Object):
             content = FileWatcherView(dialog.getFile())
             content.show()
             label.connect("close-clicked", self.onCloseTab, content)
+            self.customPages[label] = content
             self.notebook.append_page(content, label)
             self.notebook.set_current_page(self.notebook.get_n_pages()-1)
 
         dialog.destroy()
-        
+
     def onCloseTab(self, label, content):
         self.notebook.remove_page(self.notebook.page_num(content))
         content.destroy()
-        
+        del self.customPages[label]
+
     def onAboutClicked(self, menuItem):
         dialog = Gtk.MessageDialog(self.window, 0,
-                                   Gtk.MessageType.QUESTION, Gtk.ButtonsType.CLOSE);
+                                   Gtk.MessageType.QUESTION, Gtk.ButtonsType.CLOSE)
 
         dialog.set_title("About Melange")
         dialog.set_markup("""\
@@ -570,10 +511,19 @@ If you defined a hotkey for Melange, pressing it while Melange is visible it wil
         if event.keyval == Gdk.KEY_Escape:
             self.window.hide()
 
-    def onDelete(self, widget, event=None):
+    def onDelete(self, widget=None, event=None):
+        tmpPages = self.customPages.copy()
+        for label, content in tmpPages.items():
+            self.onCloseTab(label, content)
         Gtk.main_quit()
         return False
 
+    def onWindowState(self, widget, event):
+        if event.new_window_state & Gdk.WindowState.ICONIFIED:
+            self._minimized = True
+        else:
+            self._minimized = False
+
     def onPickerClicked(self, widget):
         self.lookingGlassProxy.StartInspector()
         self.window.hide()
@@ -585,7 +535,7 @@ If you defined a hotkey for Melange, pressing it while Melange is visible it wil
     def createPage(self, text, moduleName):
         module = __import__("page_%s" % moduleName)
         module.lookingGlassProxy = self.lookingGlassProxy
-        module.cinnamonLog = self
+        module.melangeApp = self
         label = Gtk.Label(text)
         page = module.ModulePage(self)
         self.pages[moduleName] = page
@@ -596,13 +546,12 @@ If you defined a hotkey for Melange, pressing it while Melange is visible it wil
         self.notebook.set_current_page(page)
 
 if __name__ == "__main__":
-    GObject.type_register(ResizeGrip)
     DBusGMainLoop(set_as_default=True)
 
     sessionBus = dbus.SessionBus ()
     request = sessionBus.request_name(MELANGE_DBUS_NAME, dbus.bus.NAME_FLAG_DO_NOT_QUEUE)
     if request != dbus.bus.REQUEST_NAME_REPLY_EXISTS:
-        app = CinnamonLog()
+        app = MelangeApp()
     else:
         object = sessionBus.get_object(MELANGE_DBUS_NAME, MELANGE_DBUS_PATH)
         app = dbus.Interface(object, MELANGE_DBUS_NAME)
